//+------------------------------------------------------------------+
//|                                      Binary file unscrambler.mq4 |
//+------------------------------------------------------------------+
// This is a script; place it in the ...\(MT4)\experts\scripts folder

#property copyright "Copyright  2012, David Louisson"
#property link      "http://www.forexfactory.com/hanover"

#property show_inputs

#include <hanover --- function header.mqh>

extern  string    FileName             = "EURUSD1_0.FXT";
extern  string    DescriptorFileType   = "FXT";
extern  string    OutputFile           = "FileView.TXT";
extern  int       StartRecord          = 1;
extern  int       NumberRecords        = 20;

string   ss, name, type, mask, hex, out, HeaderDesc, RecordDesc, arr[600];
int      leng, off, tot, hd, hi, ho, ival, HeaderLength, RecordLength, FileSizeBytes;
double   dval, DataRecords;


//+------------------------------------------------------------------+
int start()  {
//+------------------------------------------------------------------+

  // Get descriptor strings.....
  hd = FileOpen(DescriptorFileType+".des",FILE_CSV|FILE_READ,'~');           if (hd<0)   return(-1);
  HeaderDesc = FileReadString(hd);
  RecordDesc = FileReadString(hd);
  FileClose(hd);

  // Process header record.....
  if (StringLen(OutputFile) < 1)
    OutputFile = "Unscrambled " + FileName;
  ho = FileOpen(OutputFile,FILE_CSV|FILE_WRITE,'~');    if (ho<0)   return(-1);
  hi = FileOpen(FileName,FILE_BIN|FILE_READ);           if (hi<0)   return(-1);
  ss = "Unscrambled view of binary file: " + FileName;
  FileWrite(ho, ss + "\n" + StringRepeat("=",StringLen(ss)) + "\n");
  FileWrite(ho,"   HEADER        Offset-decimal   Length   Field name             Type         0  1  2  3  4  5  6  7     Value");
  FileWrite(ho,"-------------    --------------   ------   --------------------   --------    ------------------------    ----------------------");
  tot = StrToStringArray(HeaderDesc,arr,";");
  main_process();
  HeaderLength = FileTell(hi);

  // Ascertain data record length from descriptor string.....  
  tot = StrToStringArray(RecordDesc,arr,";");
  RecordLength = 0;
  for (int i=0; i<tot; i+=3)   {
    type = ExtractAlpha(StringUpper(arr[i+1]),"A");
    leng = StrToNumber(arr[i+1]);
    if (type=="T" || type=="C")   leng = 4;
    if (type=="B")                leng = 1;
    RecordLength += leng;    
  }
  
  // Output file and record stats.....
  FileSizeBytes = FileSize(hi);
  DataRecords   = DivZero(FileSizeBytes - HeaderLength, RecordLength);
  FileWrite(ho,NumberToStr(HeaderLength,"'\n\nHeader record length   = ',12'  bytes'"));
  FileWrite(ho,NumberToStr(RecordLength,  "'Data record length     = ',12'  bytes'"));
  FileWrite(ho,NumberToStr(FileSizeBytes, "'File size              = ',12'  bytes'"));
  FileWrite(ho,NumberToStr(DataRecords,   "'Data records in file   = ',12'\n'"));
  
  // Process data records.......
  for (int rec=StartRecord; rec<StartRecord+NumberRecords; rec++)   {
    FileSeek(hi,(rec-1)*RecordLength+HeaderLength,SEEK_SET);
    FileWrite(ho,"\nRec# "+NumberToStr(rec,"Z8")+"    Offset-decimal   Length   Field name             Type         0  1  2  3  4  5  6  7     Value");
    FileWrite(ho,"-------------    --------------   ------   --------------------   --------    ------------------------    ----------------------");
    main_process();
  }
  
  // Close files and end.....
  FileClose(ho);
  FileClose(hi);
  MessageBox("Output to '" + OutputFile + "' file is complete");
  return(0);
}


//+------------------------------------------------------------------+
void  main_process()  {
//+------------------------------------------------------------------+
  // Parse the relvant descriptor string (now in arr[]), and output one line for each field in the HST/FXT record.......
  for (int i=0; i<tot; i+=3)   {
    name = arr[i];                                       // first item in each triplet is the field name
    type = ExtractAlpha(StringUpper(arr[i+1]),"A");      // second item in each triplet is the type/length
    leng = StrToNumber(arr[i+1]);
    if (type=="T" || type=="C")   leng = 4;
    if (type=="B")                leng = 1;
    mask = arr[i+2];                                     // third item ine ach triplet is the override mask
    off  = FileTell(hi);                                 // save the current file pointer
    
    // build the hex values for the field.....
    hex  = "";
    for (int j=0; j<MathMin(8,leng); j++)  {
      ss = FileReadString(hi,1);
      hex = AppendIfNotNull(hex," ") + NumberToBase(StrToChar(ss),16,2);
    }
    
    // build the output line (hex & dec offsets, length, field name, type, and field's hex values.....
    out = " " + EmbedString(NumberToBase(off,16,8),"## ## ## ##") + NumberToStr(off,"'    ',12' '") + NumberToStr(leng,",6'   '");
    out = out + StrToStr(name,"L23") + StrToStr(convert_type(type),"L12") + StrToStr(hex,"L28");
    
    // now return and re-read the same field in string/integer/double/datetime/color/bool format, depending on its data type.....
    FileSeek(hi,off,SEEK_SET);  
    switch (StringFind("SIDTCB",type))    {
      case 0 :   // string
        ss = FileReadString(hi,leng);
        if (StringLen(mask)<1)   mask = "T60";                // default mask (if no override specified in descriptor string)
        out = out + StrToStr(ss, mask);
        break;
      case 1 :   // integer
        if (leng==4)       ival = FileReadInteger(hi,LONG_VALUE);           else
        if (leng==2)       ival = FileReadInteger(hi,SHORT_VALUE);          else
        if (leng==1)       ival = FileReadInteger(hi,CHAR_VALUE); 
        if (StringLen(mask)<1)   mask = "T,-12";              // default mask (if no override specified in descriptor string)
        out = out + NumberToStr(ival, mask);
        break;
      case 2 :   // double
        if (leng==8)       dval = FileReadDouble(hi,DOUBLE_VALUE);      else
        if (leng==4)       dval = FileReadDouble(hi,FLOAT_VALUE);
        if (StringLen(mask)<1)   mask = "T~,-8.5";            // default mask (if no override specified in descriptor string)
        out = out + NumberToStr(dval, mask);
        break;
      case 3 :   // datetime
        ival = FileReadInteger(hi,LONG_VALUE);
        if (StringLen(mask)<1)   mask = "w D n Y  H:I:S";     // default mask (if no override specified in descriptor string)
        out = out + NumberToStr(ival, "T,-12'    '") + DateToStr(ival, mask);
        break;
      case 4 :   // color
        ival = FileReadInteger(hi,LONG_VALUE);
        out = out + NumberToStr(ival, "T,-12'    '") + ColorToStr(ival);
        break;
      case 5 :   // boolean
        ival = FileReadInteger(hi,CHAR_VALUE);
        out = out + BoolToStr(ival);
        break;
      default :
        break;  
    }
    
    // output line is now built, write it to file.....   
    FileWrite(ho,out);
  }  
  return(0);
}


//+------------------------------------------------------------------+
string  convert_type(string type)  {
//+------------------------------------------------------------------+
  if (type=="S")     return("string");
  if (type=="I")     return("int");
  if (type=="D")     return("double");
  if (type=="T")     return("datetime");
  if (type=="C")     return("color");
  if (type=="B")     return("bool");
  return("unknown");
}

//+------------------------------------------------------------------+
#include <hanover --- extensible functions.mqh>